use std::fmt::Display;

#[allow(unused_imports)]
use crate::{query::IntoQuery, seal::Sealed};

#[allow(unused_imports)]
use super::Operable;

#[derive(Debug, Clone, Default)]
pub(crate) struct Mod<'a, M> {
    mod_type: M,
    labels: Vec<&'a str>,
}

impl<'a, T: IntoIterator<Item = &'a str>, M> From<(M, T)> for Mod<'a, M> {
    fn from((mod_type, labels): (M, T)) -> Self {
        Mod {
            mod_type,
            labels: labels.into_iter().collect(),
        }
    }
}

impl Display for Mod<'_, MatchType> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if !self.labels.is_empty() && !matches!(self.mod_type, MatchType::None) {
            write!(
                f,
                " {matchtype}({labels})",
                matchtype = self.mod_type,
                labels = self.labels.join(",")
            )?;
        }
        Ok(())
    }
}

impl Display for Mod<'_, GroupType> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if !matches!(self.mod_type, GroupType::None) {
            if self.labels.is_empty() {
                return write!(f, " {grp_type}", grp_type = self.mod_type,);
            }

            return write!(
                f,
                " {grp_type}({labels})",
                grp_type = self.mod_type,
                labels = self.labels.join(","),
            );
        }
        Ok(())
    }
}

#[derive(Default, Debug, Clone, Copy)]
pub(crate) enum MatchType {
    On,
    Ignoring,
    #[default]
    None,
}

impl Display for MatchType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            MatchType::On => "on",
            MatchType::Ignoring => "ignoring",
            MatchType::None => "",
        })
    }
}

#[derive(Debug, Clone, Copy, Default)]
pub(crate) enum GroupType {
    Left,
    Right,
    #[default]
    None,
}

impl Display for GroupType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            GroupType::Left => "group_left",
            GroupType::Right => "group_right",
            GroupType::None => "",
        })
    }
}

#[derive(Default, Debug, Clone, Copy)]
pub(crate) enum AggrType {
    By,
    Without,
    #[default]
    None,
}

impl Display for AggrType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            AggrType::By => "by",
            AggrType::Without => "without",
            AggrType::None => "",
        })
    }
}

impl Display for Mod<'_, AggrType> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if !self.labels.is_empty() {
            return write!(
                f,
                " {mod_type} ({labels})",
                mod_type = self.mod_type,
                labels = self.labels.join(",")
            );
        }
        Ok(())
    }
}

/// MetricsQL's `keep_metric_names` modifier
///
/// This modifier can only be applied using the `keep_metric_names` method made available by the [`QueryExt`](crate::query::QueryExt) trait.
#[cfg(feature = "metricsql")]
#[derive(Debug, Clone)]
pub struct KeepNames<E: Operable>(pub(crate) E);

#[cfg(feature = "metricsql")]
impl<E: Operable> Display for KeepNames<E> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{exp} keep_metric_names", exp = self.0)
    }
}

#[cfg(feature = "metricsql")]
impl<E: Operable> Sealed for KeepNames<E> {}
#[cfg(feature = "metricsql")]
impl<E: Operable> Operable for KeepNames<E> {}

#[cfg(feature = "metricsql")]
impl<E: Operable> IntoQuery for KeepNames<E> {
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

#[cfg(test)]
mod tests {
    use crate::query::{Metric, QueryExt};

    #[cfg(feature = "metricsql")]
    #[test]
    fn keep_names() {
        let qry = Metric::new("metric").keep_metric_names();
        assert_eq!(qry.to_string(), "metric keep_metric_names");
    }
}
